"""
How does this work?
When you fetch the page of a user on instagram you get an html with a javascript embedded
in it with a json object embedded in that. This json object describes the user, his id,
his profile photo and the first 12 images for that user.
If you want to more workers you have to do a follow-up AJAX request to the server.
"""
import json
from typing import List

from lxml import etree

import pyscrapers.core.utils


def scrape_instagram(user_id: str, session) -> List[str]:
    domain = 'www.instagram.com'
    base = 'https://{domain}'.format(domain=domain)

    url = '{base}/{user_id}/'.format(base=base, user_id=user_id)

    r = session.get(url)
    root = pyscrapers.core.utils.get_html_dom_content(r)
    # scrape.utils.print_element(root)

    urls = []
    # register regular expressions with lxml
    # this means that we can use regular expression functions like 'match'
    # by specifying 're:match' in our xpath expressions
    ns = etree.FunctionNamespace("http://exslt.org/regular-expressions")
    ns.prefix = 're'
    e_a = root.xpath('//script[re:match(text(), "^window._sharedData")]')
    assert len(e_a) == 1
    e_a = e_a[0]
    data = e_a.text
    json_text = data[data.find('{'):data.rfind('}') + 1]
    d = json.loads(json_text)
    my_list = d['entry_data']['ProfilePage']
    assert (len(my_list) == 1)
    c = my_list[0]["graphql"]["user"]
    if 'profile_pic_url_hd' in c:
        urls.append(c['profile_pic_url_hd'])
    elif 'profile_pic_url' in c:
        urls.append(c['profile_pic_url'])
    user_id = c['id']
    url2 = '{base}/graphql/query/'.format(base=base)
    has_next_page = True
    end_cursor = None
    while has_next_page:
        variables = {
                'id': user_id,
                'first': 12,
        }
        if end_cursor:
            variables['after'] = end_cursor
        # noinspection SpellCheckingInspection
        params = {
            # 'query_id': 17888483320059182,
            'query_hash': 'bd0d6d184eefd4d0ce7036c11ae58ed9',
            'variables': json.dumps(variables)
        }
        r2 = session.get(url2, params=params)
        # print(r2.json())
        # quit()
        data_user = r2.json()['data']['user']['edge_owner_to_timeline_media']
        has_next_page = data_user['page_info']['has_next_page']
        end_cursor = data_user['page_info']['end_cursor']
        for node in data_user['edges']:
            # if it is a video, then append the video, too
            if node['node']['is_video']:
                urls.append(node['node']['video_url'])
                # print(json.dumps(node['node'], indent=4, sort_keys=True))
            urls.append(node['node']['display_url'])
    return urls



"""
Module to handle scraping of pornhub.

References:
- https://pypi.org/project/pornhub-api/
"""
import os
from itertools import islice

import pornhub_api
from pornhub_api import PornhubApi

from pyscrapers.configs import ConfigPornhubSearch, ConfigPornhubDownload
from pyscrapers.workers.youtube_dl_handlers import youtube_dl_download_urls


def print_stars_all(api: pornhub_api.api.PornhubApi) -> None:
    stars_results = api.stars.all()
    for star in islice(stars_results.stars, 0, ConfigPornhubSearch.limit):
        print(star.star.star_name)


def print_categories(api: pornhub_api.api.PornhubApi) -> None:
    categories = api.video.categories()
    for category in categories.categories:
        print("id [{}], name [{}]".format(category.id, category.category))


def print_tags(api: pornhub_api.api.PornhubApi) -> None:
    tags = api.video.tags(ConfigPornhubSearch.literal)
    for tag in tags:
        print("tag [{}]".format(tag))


def download() -> None:
    """
    Download movies from pornhub
    """
    api = PornhubApi()

    limit = ConfigPornhubSearch.limit
    over = False
    page = 1
    counter = 0
    while not over:
        try:
            kwargs = dict()
            if ConfigPornhubSearch.use_ordering:
                kwargs["ordering"] = ConfigPornhubSearch.ordering
            if ConfigPornhubSearch.use_period:
                kwargs["period"] = ConfigPornhubSearch.period
            if ConfigPornhubSearch.use_tags:
                kwargs["tags"] = ConfigPornhubSearch.tags
            data = api.search.search(
                ConfigPornhubSearch.query,
                page=page,
                **kwargs,
            )
            urls = [video.url for video in data.videos]
            if limit is not None:
                urls = list(islice(urls, 0, limit-counter))
            youtube_dl_download_urls(
                urls,
                os.path.join(
                    ConfigPornhubDownload.folder,
                    ConfigPornhubSearch.query,
                )
            )
            counter += len(urls)
            page += 1
            if counter == limit:
                break
        except ValueError as e:
            code = e.args[0]["code"]
            if code == "2001":  # no videos found (end of results)
                over = True
            else:
                raise e



"""
The default group of operations that pyscrapers has
"""
import logging
import shelve

import requests
from pytconf.config import register_endpoint, register_function_group

import pyscrapers.core.utils
import pyscrapers.version
from pyscrapers.configs import ConfigDebugRequests, ConfigCookiesSource, ConfigSiteId, ConfigPornhubSearch, \
    ConfigYoutubeDl
from pyscrapers.workers.drumeo import get_number_of_pages, get_courses, get_course_details, get_course_urls, \
    download_course
from pyscrapers.workers.facebook import scrape_facebook
from pyscrapers.workers.instagram import scrape_instagram
from pyscrapers.workers.mambaru import scrape_mambaru
from pyscrapers.workers.pornhub import download
from pyscrapers.workers.travelgirls import scrape_travelgirls
from pyscrapers.workers.vk import scrape_vk
from pyscrapers.workers.youtube_dl_handlers import youtube_dl_handler

GROUP_NAME_DEFAULT = "default"
GROUP_DESCRIPTION_DEFAULT = "all pyscapers commands"


def register_group_default():
    """
    register the name and description of this group
    """
    register_function_group(
        function_group_name=GROUP_NAME_DEFAULT,
        function_group_description=GROUP_DESCRIPTION_DEFAULT,
    )


@register_endpoint(
    configs=[],
    suggest_configs=[],
    group=GROUP_NAME_DEFAULT,
)
def version() -> None:
    """
    Print version
    """
    print(pyscrapers.version.VERSION_STR)


@register_endpoint(
    configs=[
        ConfigSiteId,
        ConfigDebugRequests,
        ConfigCookiesSource,
    ],
    suggest_configs=[],
    group=GROUP_NAME_DEFAULT,
)
def photos():
    """
    Download photo albums from various sites
    """
    if ConfigDebugRequests.debug:
        pyscrapers.core.utils.debug_requests()
    ConfigCookiesSource.config_cookies()
    session = requests.Session()
    session.cookies = ConfigCookiesSource.cookies
    urls = []
    if ConfigSiteId.site == "facebook":
        urls = scrape_facebook(ConfigSiteId.user_id, session)
    if ConfigSiteId.site == "instagram":
        urls = scrape_instagram(ConfigSiteId.user_id, session)
    if ConfigSiteId.site == "travelgirls":
        urls = scrape_travelgirls(ConfigSiteId.user_id, session)
    if ConfigSiteId.site == "vk":
        urls = scrape_vk(ConfigSiteId.user_id, session)
    if ConfigSiteId.site == "mamba.ru":
        urls = scrape_mambaru(ConfigSiteId.user_id, session)
    pyscrapers.core.utils.download_urls(session, urls, start=ConfigSiteId.start)
    session.close()


@register_endpoint(
    configs=[
        ConfigDebugRequests,
        ConfigCookiesSource,
    ],
    suggest_configs=[],
    group=GROUP_NAME_DEFAULT,
)
def drumeo():
    """
    Download videos from drumeo
    """
    if ConfigDebugRequests.debug:
        pyscrapers.core.utils.debug_requests()

    ConfigCookiesSource.config_cookies()

    session = requests.Session()
    session.cookies = ConfigCookiesSource.cookies

    logger = logging.getLogger(__name__)
    courses = False
    reload = {}
    with shelve.open("cache.db") as d:
        if "courses" in d:
            list_of_courses = d["courses"]
            print("got from cache [{}] courses".format(len(list_of_courses)))
        else:
            pages = get_number_of_pages(courses=courses, session=session)
            print("number of pages is [{}]".format(pages))
            list_of_courses = get_courses(pages, courses=courses, session=session)
            print("got [{}] courses".format(len(list_of_courses)))
            d["courses"] = list_of_courses
        for i, course in enumerate(list_of_courses):
            logger.info("course number [%s]", i)
            if course.number in d and course.number not in reload:
                list_of_courses[i] = d[course.number]
                logger.info("got from cache [%s]", list_of_courses[i])
            else:
                get_course_details(course, courses=courses, session=session)
                get_course_urls(course, courses=courses, session=session)
                print(course)
                d[course.number] = course
            download_course(list_of_courses[i], session)
    session.close()


@register_endpoint(
    configs=[
        ConfigPornhubSearch,
    ],
    suggest_configs=[],
    group=GROUP_NAME_DEFAULT,
)
def pornhub():
    """
    Download movies from pornhub
    """
    download()


@register_endpoint(
    configs=[
        ConfigYoutubeDl,
    ],
    suggest_configs=[],
    group=GROUP_NAME_DEFAULT,
)
def youtube_dl():
    """
    Download movies from pornhub
    """
    youtube_dl_handler()



"""
Module that handles the interaction with the youtube_dl library

References:
- https://github.com/ytdl-org/youtube-dl/blob/master/README.md#embedding-youtube-dl
"""
import os
from typing import List

import youtube_dl

from pyscrapers.configs import ConfigYoutubeDl


class MyLogger(object):
    def debug(self, msg):
        pass

    def warning(self, msg):
        pass

    # noinspection PyMethodMayBeStatic
    def error(self, msg):
        print(msg)


def youtube_dl_handler() -> None:
    youtube_dl_download_url(ConfigYoutubeDl.url, ConfigYoutubeDl.folder)


def youtube_dl_download_url(url: str, folder: str) -> None:
    youtube_dl_download_urls([url], folder)


def youtube_dl_download_urls(urls: List[str], folder: str) -> None:
    ydl_opts = {
        'format': 'bestaudio/best',
        # 'logger': MyLogger(),
        'outtmpl': os.path.join(folder, '%(title)s-%(id)s.%(ext)s'),
    }
    with youtube_dl.YoutubeDL(ydl_opts) as ydl:
        ydl.download(urls)



import json
from typing import List

from pyscrapers.core.utils import get_http_status_string


def scrape_mambaru(user_id: str, session) -> List[str]:
    # raise ValueError("mamba still not implemented")
    # main_url = 'https://www.mamba.ru/{user_id}'.format(user_id=user_id)
    main_url = 'https://www.mamba.ru/mobile/api/v5.17.0.0/?reqType=json'
    request_obj = {"langId": "en", "dateType": "timestamp", "limit": 10000, "sysRequestsContainer": [
            {"uri": "/users/{}/albums/workers/".format(user_id), "method": "GET", "params":
                {"langId": "en", "dateType": "timestamp", "limit": -1}
             }
        ]
    }
    urls = []
    response = session.post(main_url, json=request_obj)
    assert response.status_code == 200, get_http_status_string(response.status_code)
    response_str = response.content.decode()
    response_obj = json.loads(response_str)
    response_obj = response_obj["sysResponsesContainer"][0]
    for album in response_obj["albums"]:
        for photo in album["workers"]:
            urls.append(photo["hugePhotoUrl"])
    return urls



"""
download photos from facebook
"""


import logging
from typing import List

import pyscrapers.core.utils


def scrape_facebook(user_id: str, session) -> List[str]:
    """
    download photos from facebook
    :param user_id:
    :param session:
    :return:
    """
    logger = logging.getLogger(__name__)

    url = 'https://www.facebook.com/{user_id}/workers'.format(user_id=user_id)
    logger.debug('url is [%s]', url)
    result = session.get(url)
    root = pyscrapers.core.utils.get_html_dom_content(result)

    urls = []
    elements_img = root.xpath('//img')
    urls.append(elements_img)
    return urls



"""
Download course material from drumeo
"""

import json
import logging
import os

import lxml.html

from pyscrapers.core.utils import download_url, download_video_if_wider

# this means that we can use regular expression functions like 'match'
# by specifying 're:match' in our xpath expressions
ns = lxml.etree.FunctionNamespace("http://exslt.org/regular-expressions")
ns.prefix = 're'


def get_number_of_pages(courses: bool, session) -> int:
    """
    Get the number of pages for all courses or pages
    :param courses:
    :param session:
    :return:
    """
    if courses:
        url = "https://www.drumeo.com/laravel/public/members-area/json/lesson-group/courses?page=1"
    else:
        url = "https://www.drumeo.com/laravel/public/members-area/json/lesson-group/library?page=1"
    result = session.get(url)
    assert result.status_code == 200
    content = result.content.decode()
    obj = json.loads(content)
    d_page_count = obj["pageCount"]
    return d_page_count


class Course:
    """
    This is an object representing one course
    """
    def __init__(self):
        """
        constructor
        """
        self.number = 0
        self.instructor = None
        self.name = None
        self.diff = None
        self.lessons = []
        self.resources = None
        self.videos = []

    def __repr__(self):
        """
        nice representation of this object
        :return:
        """
        return ",".join([
            self.number, self.name, self.instructor, self.diff,
            str(self.lessons), str(self.resources),
            str(self.videos)])

    def add_lesson(self, lesson):
        """
        Add a lesson to the course
        :param lesson:
        :return:
        """
        self.lessons.append(lesson)

    def add_video(self, video, quality):
        """
        Add a video to the course
        :param video:
        :param quality:
        :return:
        """
        self.videos.append((video, quality))


def get_url(courses: bool, page: int) -> str:
    if courses:
        url = "https://www.drumeo.com/laravel/public/members-area/json/lesson-group/courses?page={}".format(page)
    else:
        url = "https://www.drumeo.com/laravel/public/members-area/json/lesson-group/library?page={}".format(page)
    return url


def get_members_url(courses: bool, number: int) -> str:
    if courses:
        url = "https://www.drumeo.com/members/lessons/courses/{}".format(number)
    else:
        url = "https://www.drumeo.com/members/lessons/library/{}".format(number)
    return url


def get_link_re(courses: bool) -> str:
    if courses:
        link_re = r"https://www.drumeo.com/laravel/public/members/lessons/courses/\d+"
    else:
        link_re = r"https://www.drumeo.com/laravel/public/members/lessons/library/\d+"
    return link_re


def get_courses(pages, courses: bool, session):
    """
    Download the list of all the courses
    :param pages:
    :param courses:
    :param session:
    :return:
    """
    collected_courses = []
    for page in range(1, pages + 1):
        result = session.get(get_url(courses, page))
        assert result.status_code == 200
        for lesson_list in json.loads(result.content.decode())["lessonsHtml"]:
            course = Course()
            root = lxml.html.fromstring(lesson_list)
            links = root.xpath('//a[re:match(@href,"{}")]'.format(get_link_re(courses)))
            assert len(links) == 2, len(links)
            course.number = links[0].get('href').split("/")[-1]
            titles = root.xpath('//h2[@class="card-title"]')
            assert len(titles) == 1
            course.name = titles[0].text
            instructors = root.xpath('//p[@class="card-sub-title card-instructor"]')
            assert len(instructors) == 1
            course.instructor = instructors[0].text
            diffs = root.xpath('//h3[@class="card-difficulty"]')
            assert len(diffs) == 1
            course.diff = diffs[0].text.strip()
            collected_courses.append(course)
    return collected_courses


def get_course_details(course: Course, courses: bool, session):
    """
    Populate the Course type object
    :param course:
    :param courses:
    :param session:
    :return:
    """
    result = session.get(get_members_url(courses, course.number))
    assert result.status_code == 200
    content = result.content.decode()
    root = lxml.html.fromstring(content)
    if courses:
        class_text = "course-lesson"
    else:
        class_text = "event-toggle download-lesson"
    lessons = root.xpath('//a[@class="{}"]'.format(class_text))
    for lesson in lessons:
        lesson_num = lesson.get('href').split('/')[-1]
        course.add_lesson(lesson_num)
    resources = root.xpath('//a[re:match(text(), "All Course Resources")]')
    if len(resources) == 1:
        resource = resources[0].get('href')
        course.resources = "http:"+resource
    if not courses:
        get_videos(root, course, session)


def get_course_urls(course, courses: bool, session):
    if not courses:
        return
    logger = logging.getLogger(__name__)
    logger.info("doing course [%s]", course)
    for lesson in course.lessons:
        if courses:
            url = "https://www.drumeo.com/members/lessons/courses/{}".format(lesson)
        else:
            url = "https://www.drumeo.com/members/lessons/library/{}".format(lesson)
        result = session.get(url)
        assert result.status_code == 200
        content = result.content.decode()
        print(content)
        root = lxml.html.fromstring(content)
        get_videos(root, course, session)


def get_videos(root, course, session):
    """
    Get all the videos pertaining to a certain course
    :param root:
    :param course:
    :param session:
    :return:
    """
    logger = logging.getLogger(__name__)
    videos = root.xpath('//div[@data-video-load-url]')
    for video in videos:
        video_url = video.get('data-video-load-url')
        logger.info("url for video info is [%s]", video_url)
        result = session.get(video_url)
        assert result.status_code == 200
        content = result.content.decode()
        data = json.loads(content)
        if 'error' in data:
            logger.info("error [%s]", data['error'])
            raise ValueError("errors, try later")
        if 'video-quality-urls' not in data:
            logger.info("did not find video-quality-urls")
            return
        video_urls = data['video-quality-urls']
        quality_numbers = sorted([int(x) for x in video_urls.keys()])
        best_vid_key = str(quality_numbers[-1])
        # print(quality_numbers, best_vid_key)
        best_vid = video_urls[best_vid_key]
        course.add_video(best_vid, best_vid_key)


def download_course(course, session):
    """
    Download a course
    :param course:
    :param session:
    :return:
    """
    folder_name = os.path.join("drumeo", course.number)
    if not os.path.isdir(folder_name):
        os.makedirs(folder_name)
    details = os.path.join(folder_name, "details.txt")
    if not os.path.isfile(details):
        with open(details, "wt") as file_handle:
            print("course_number: {}".format(course.number), file=file_handle)
            print("course_name: {}".format(course.name), file=file_handle)
            print("course_difficulty: {}".format(course.diff), file=file_handle)
            print("instructor: {}".format(course.instructor), file=file_handle)
    if course.resources is not None:
        download_url(course.resources, os.path.join(folder_name, "resources.zip"), session)
    for i, (video, quality) in enumerate(course.videos):
        download_video_if_wider(session, video, os.path.join(folder_name, "{}.mp4".format(i)), width=int(quality))
